# Copyright (c) Meta Platforms, Inc. and affiliates.

# pyre-strict

from __future__ import annotations

import ast
import builtins
import sys
from ast import (
    Assign,
    AST,
    AsyncFunctionDef,
    Attribute,
    BinOp,
    BoolOp,
    Call,
    ClassDef,
    cmpop,
    Compare,
    Constant,
    DictComp,
    expr,
    FunctionDef,
    Module,
    Name,
    UnaryOp,
)
from collections.abc import Callable as typingCallable, Generator
from contextlib import contextmanager
from types import CodeType
from typing import Any, cast

from .. import consts, opcode_static
from ..opcodebase import Opcode
from ..pyassem import (
    Block,
    IndexedSet,
    PyFlowGraph,
    PyFlowGraph312,
    PyFlowGraph314,
    PyFlowGraph315,
    PyFlowGraphCinder310,
)
from ..pycodegen import (
    CinderCodeGenerator310,
    CinderCodeGenerator312,
    CinderCodeGenerator314,
    CinderCodeGenerator315,
    CodeGenerator,
    CodeGenTree,
    compile_code,
    CompNode,
    FuncOrLambda,
    PatternContext,
)
from ..strict import (
    StrictCodeGenerator310,
    StrictCodeGenerator312,
    StrictCodeGenerator314,
)
from ..strict.code_gen_base import StrictCodeGenBase
from ..strict.common import FIXED_MODULES
from ..symbols import BaseSymbolVisitor, ClassScope, ModuleScope, Scope
from .compiler import Compiler
from .definite_assignment_checker import DefiniteAssignmentVisitor
from .effects import NarrowingEffect, TypeState
from .module_table import ModuleFlag, ModuleTable
from .type_binder import UsedRefinementField
from .types import (
    ASYNC_CACHED_PROPERTY_IMPL_PREFIX,
    AsyncCachedPropertyMethod,
    AwaitableType,
    CACHED_PROPERTY_IMPL_PREFIX,
    CachedPropertyMethod,
    CInstance,
    Class,
    CType,
    Dataclass,
    DataclassDecorator,
    DataclassField,
    DecoratedMethod,
    DynamicInstance,
    EAGER_IMPORT_NAME,
    Function,
    FunctionContainer,
    GenericClass,
    KnownBoolean,
    Slot,
    TMP_VAR_PREFIX,
    TType,
    TypeDescr,
    Value,
)

try:
    from cinder import _set_qualname
except ImportError:

    def _set_qualname(code: CodeType, qualname: str) -> None:
        pass


def exec_static(
    source: str,
    locals: dict[str, object],
    globals: dict[str, object],
    modname: str = "<module>",
) -> None:
    code = compile_code(
        source, "<module>", "exec", compiler=StaticCodeGenerator, modname=modname
    )
    if "<fixed-modules>" not in globals:
        globals["<fixed-modules>"] = FIXED_MODULES
    if "<builtins>" not in globals:
        globals["<builtins>"] = builtins.__dict__

    exec(code, locals, globals)


class StaticPatternContext(PatternContext):
    def __init__(self) -> None:
        super().__init__()
        self.type_dict: dict[str, Class] = {}

    def clone(self) -> StaticPatternContext:
        pc = StaticPatternContext()
        pc.stores = list(self.stores)
        pc.allow_irrefutable = self.allow_irrefutable
        pc.fail_pop = list(self.fail_pop)
        pc.on_top = self.on_top
        pc.type_dict = self.type_dict.copy()
        return pc


class PyFlowGraphStatic310(PyFlowGraphCinder310):
    opcode: Opcode = opcode_static.opcode


class PyFlowGraphStatic312(PyFlowGraph312):
    opcode: Opcode = opcode_static.opcode


class InitSubClassGenerator:
    def __init__(self, flow_graph: PyFlowGraph, qualname: str) -> None:
        self.flow_graph = flow_graph
        self.qualname = qualname

    @property
    def scope(self) -> Scope:
        raise NotImplementedError()

    @property
    def name(self) -> str:
        return self.qualname

    def getCode(self) -> CodeType:
        code = self.flow_graph.getCode()
        _set_qualname(code, self.qualname)
        return code


class StaticCodeGenBase(StrictCodeGenBase):
    flow_graph = PyFlowGraph
    _default_cache: dict[type[ast.AST], typingCallable[..., None]] = {}
    pattern_context = StaticPatternContext
    # Defined in subclasses; this is an explicit receiver class for
    # self.defaultCall() and self.make_child_codegen to dispatch to
    parent_impl: type[CodeGenerator] = CodeGenerator

    def __init__(
        self,
        parent: CodeGenerator | None,
        node: AST,
        symbols: BaseSymbolVisitor,
        graph: PyFlowGraph,
        compiler: Compiler,
        modname: str,
        flags: int = 0,
        optimization_lvl: int = 0,
        enable_patching: bool = False,
        builtins: dict[str, Any] = builtins.__dict__,
        future_flags: int | None = None,
    ) -> None:
        super().__init__(
            parent,
            node,
            symbols,
            graph,
            flags=flags,
            optimization_lvl=optimization_lvl,
            builtins=builtins,
            future_flags=future_flags,
        )
        self.compiler = compiler
        self.modname = modname
        # Use this counter to allocate temporaries for loop indices
        self._tmpvar_loopidx_count = 0
        self.cur_mod: ModuleTable = self.compiler.modules[modname]
        self.enable_patching = enable_patching
        self.scope_to_node: dict[Scope, AST] = {
            scope: node for node, scope in self.scopes.items()
        }

    def _is_static_compiler_disabled(self, node: AST) -> bool:
        if not isinstance(node, (AsyncFunctionDef, FunctionDef, ClassDef)):
            # Static compilation can only be disabled for functions and classes.
            return False
        if node in self.cur_mod.compile_non_static:
            return True
        scope_node = self.scope_to_node[self.scope]
        fn = None
        if isinstance(scope_node, ClassDef):
            klass = self.get_type(scope_node)
            if klass:
                assert isinstance(klass, Class)
                if klass.donotcompile:
                    # If static compilation is disabled on the entire class, it's skipped for all contained
                    # methods too.
                    return True
                fn = klass.get_own_member(node.name)

        if fn is None:
            # Wasn't a method, let's check if it's a module level function
            fn = self.get_type(node)

        if isinstance(fn, (Function, DecoratedMethod)):
            return fn.donotcompile

        return False

    def _is_type_checked(self, t: Class) -> bool:
        return t is not self.compiler.type_env.object and t.is_nominal_type

    def _get_arg_types(
        self,
        func: FuncOrLambda | CompNode,
        args: ast.arguments,
        graph: PyFlowGraph,
    ) -> tuple[object, ...]:
        arg_checks = []
        cellvars = graph.cellvars
        is_comprehension = not isinstance(
            func, (ast.AsyncFunctionDef, ast.FunctionDef, ast.Lambda)
        )

        for i, arg in enumerate(args.posonlyargs):
            t = self.get_type(arg)
            if self._is_type_checked(t.klass):
                arg_checks.append(self._calculate_idx(arg.arg, i, cellvars))
                arg_checks.append(t.klass.type_descr)

        for i, arg in enumerate(args.args):
            # Comprehension nodes don't have arguments when they're typed; make
            # up for that here.
            t = (
                self.compiler.type_env.DYNAMIC
                if is_comprehension
                else self.get_type(arg)
            )
            if self._is_type_checked(t.klass):
                arg_checks.append(
                    self._calculate_idx(arg.arg, i + len(args.posonlyargs), cellvars)
                )
                arg_checks.append(t.klass.type_descr)

        for i, arg in enumerate(args.kwonlyargs):
            t = self.get_type(arg)
            if self._is_type_checked(t.klass):
                arg_checks.append(
                    self._calculate_idx(
                        arg.arg,
                        i + len(args.posonlyargs) + len(args.args),
                        cellvars,
                    )
                )
                arg_checks.append(t.klass.type_descr)

        # we should never emit arg checks for object
        assert not any(td == ("builtins", "object") for td in arg_checks[1::2])

        return tuple(arg_checks)

    def get_type(self, node: AST) -> Value:
        return self.cur_mod.types[node]

    def get_node_data(self, key: AST, data_type: type[TType]) -> TType:
        return self.cur_mod.get_node_data(key, data_type)

    def get_opt_node_data(self, key: AST, data_type: type[TType]) -> TType | None:
        return self.cur_mod.get_opt_node_data(key, data_type)

    def set_node_data(self, key: AST, data_type: type[TType], value: TType) -> None:
        self.cur_mod.set_node_data(key, data_type, value)

    @classmethod
    def make_code_gen(
        cls,
        module_name: str,
        tree: AST,
        filename: str,
        source: str | bytes | ast.Module | ast.Expression | ast.Interactive,
        flags: int,
        optimize: int,
        ast_optimizer_enabled: bool = True,
        enable_patching: bool = False,
        builtins: dict[str, Any] = builtins.__dict__,
    ) -> StaticCodeGenBase:
        assert ast_optimizer_enabled

        compiler = Compiler(cls)
        code_gen = compiler.code_gen(
            module_name, filename, tree, source, optimize, enable_patching, builtins
        )
        return code_gen

    def make_child_codegen(
        self,
        tree: CodeGenTree,
        graph: PyFlowGraph,
        name: str | None = None,
    ) -> CodeGenerator:
        if self._is_static_compiler_disabled(tree):
            return self.parent_impl(
                self,
                tree,
                self.symbols,
                graph,
                flags=self.flags,
                optimization_lvl=self.optimization_lvl,
            )
        return self.__class__(
            self,
            tree,
            self.symbols,
            graph,
            compiler=self.compiler,
            modname=self.modname,
            optimization_lvl=self.optimization_lvl,
            enable_patching=self.enable_patching,
        )

    def make_function_graph(
        self,
        func: FuncOrLambda | CompNode,
        func_args: ast.arguments,
        filename: str,
        scopes: dict[AST, Scope],
        class_name: str | None,
        name: str,
        first_lineno: int,
    ) -> PyFlowGraph:
        graph = super().make_function_graph(
            func, func_args, filename, scopes, class_name, name, first_lineno
        )

        if self._is_static_compiler_disabled(func):
            return graph

        graph.setFlag(consts.CI_CO_STATICALLY_COMPILED)
        arg_types = self._get_arg_types(func, func_args, graph)
        ret_type = self._get_return_type(func)
        graph.extra_consts.append((arg_types, ret_type.type_descr))
        return graph

    def _get_return_type(self, func: FuncOrLambda | CompNode) -> Class:
        if isinstance(func, (FunctionDef, AsyncFunctionDef)):
            function = self.get_func_container(func)
            klass = function.return_type.resolved()
            if not self._is_type_checked(klass):
                return self.compiler.type_env.object
        else:
            klass = self.get_type(func).klass
        if isinstance(klass, AwaitableType):
            klass = klass.type_args[0]
        return klass

    @contextmanager
    def new_loopidx(self) -> Generator[str, None, None]:
        self._tmpvar_loopidx_count += 1
        try:
            yield f"{TMP_VAR_PREFIX}.{self._tmpvar_loopidx_count}"
        finally:
            self._tmpvar_loopidx_count -= 1

    def _resolve_class(self, node: ClassDef) -> Class | None:
        klass = self.get_type(node)
        if not isinstance(klass, Class) or klass is self.compiler.type_env.dynamic:
            return
        return klass

    def emit_build_class(
        self,
        node: ast.ClassDef,
        class_body: CodeGenerator,
        outer_scope: CodeGenerator | None = None,
    ) -> None:
        if (outer_scope is not None and outer_scope is not self) or getattr(
            node, "type_params", None
        ):
            # TASK(T210915794): We don't support type params, we need to pass the
            # .generic_base as part of the bases
            raise NotImplementedError("Type params not supported in static Python yet")

        klass = self._resolve_class(node)
        if not isinstance(self.scope, ModuleScope) or klass is None:
            # If a class isn't top-level then it's not going to be using static
            # features (we could relax this in the future for classes nested in classes).
            return super().emit_build_class(node, class_body)

        self.emit_closure(class_body, 0)
        self.emit("LOAD_CONST", node.name)

        if node.keywords:
            for keyword in node.keywords:
                self.emit("LOAD_CONST", keyword.arg)
                self.visit(keyword.value)

            self.emit("BUILD_MAP", len(node.keywords))
        else:
            self.emit("LOAD_CONST", None)

        class_scope = class_body.scope
        assert isinstance(class_scope, ClassScope)
        self.emit("LOAD_CONST", class_scope.needs_class_closure)

        assert klass is not None
        final_methods: list[str] = []
        for class_or_subclass in klass.mro:
            final_methods.extend(class_or_subclass.get_own_final_method_names())
        self.emit("LOAD_CONST", tuple(sorted(final_methods)))

        if klass.is_final:
            self.emit("LOAD_CONST", True)
        else:
            self.emit("LOAD_CONST", False)

        count = 0
        for name, value in klass.members.items():
            if isinstance(value, CachedPropertyMethod):
                self.emit("LOAD_CONST", CACHED_PROPERTY_IMPL_PREFIX + name)
                count += 1
            elif isinstance(value, AsyncCachedPropertyMethod):
                self.emit("LOAD_CONST", ASYNC_CACHED_PROPERTY_IMPL_PREFIX + name)
                count += 1

        self.emit("BUILD_TUPLE", count)

        for base in node.bases:
            self.visit(base)

        self.emit(
            "INVOKE_FUNCTION",
            ((("_static",), "__build_cinder_class__"), 7 + len(node.bases)),
        )

    def processBody(
        self, node: FuncOrLambda, body: list[ast.stmt], gen: CodeGenerator
    ) -> None:
        if isinstance(node, (ast.FunctionDef | ast.AsyncFunctionDef)):
            # check for any unassigned primitive values and force them to be
            # assigned.
            visitor = DefiniteAssignmentVisitor(self.symbols.scopes[node])
            visitor.analyzeFunction(node)
            for unassigned in visitor.unassigned:
                node_type = self.get_type(unassigned)
                if isinstance(node_type, CInstance):
                    assert isinstance(gen, StaticCodeGenBase)
                    node_type.emit_init(unassigned, gen)

        super().processBody(node, body, gen)

    def visitFunctionOrLambda(
        self, node: ast.FunctionDef | ast.AsyncFunctionDef | ast.Lambda
    ) -> None:
        if isinstance(node, ast.Lambda):
            return super().visitFunctionOrLambda(node)

        function = self.get_func_container(node)
        name = function.emit_function(node, self)
        self.storeName(name)

    def get_func_container(self, node: ast.AST) -> FunctionContainer:
        function = self.get_type(node)
        if not isinstance(function, FunctionContainer):
            raise RuntimeError("bad value for function")

        return function

    def walkClassBody(self, node: ClassDef, gen: CodeGenerator) -> None:
        klass = self._resolve_class(node)
        super().walkClassBody(node, gen)
        if not klass:
            return

        if not klass.has_init_subclass:
            gen_scope = gen.scope
            assert isinstance(gen_scope, ClassScope)

            # we define a custom override of __init_subclass__
            if not gen_scope.needs_class_closure:
                # we need to call super(), which will need to have
                # __class__ available if it's not already...
                gen.graph.cellvars.get_index("__class__")
                gen_scope.cells["__class__"] = 1
                gen_scope.needs_class_closure = True

            init_graph = self.flow_graph(
                "__init_subclass__",
                self.graph.filename,
                None,
                0,
                ("cls",),
                (),
                (),
                optimized=1,
                docstring=None,
            )
            init_graph.freevars.get_index("__class__")
            init_graph.emit_prologue()
            init_graph.emit("LOAD_GLOBAL", "super")
            init_graph.emit("LOAD_DEREF", "__class__")
            init_graph.emit("LOAD_FAST", "cls")
            init_graph.emit_super_call("__init_subclass__", True)
            init_graph.emit_call_method(0)
            init_graph.emit("POP_TOP")

            init_graph.emit("LOAD_FAST", "cls")
            init_graph.emit("INVOKE_FUNCTION", ((("_static",), "init_subclass"), 1))
            init_graph.emit("RETURN_VALUE")

            gen.emit("LOAD_CLOSURE", "__class__")
            gen.emit("BUILD_TUPLE", 1)
            init_gen = InitSubClassGenerator(
                init_graph, self.get_qual_prefix(gen) + ".__init_subclass__"
            )
            gen.emit_make_function(
                init_gen, self.get_qual_prefix(gen) + ".__init_subclass__", 8
            )
            gen.emit("STORE_NAME", "__init_subclass__")

        assert isinstance(gen, StaticCodeGenBase)
        klass.emit_extra_members(node, gen)

        class_mems_with_overrides = [
            name
            for name, value in klass.members.items()
            if (isinstance(value, Slot) and not value.is_classvar)
            or isinstance(value, CachedPropertyMethod)
            or isinstance(value, AsyncCachedPropertyMethod)
        ]

        class_mems = [
            name
            for name in class_mems_with_overrides
            if not isinstance(mem := klass.members[name], Slot) or not mem.override
        ]

        if klass.allow_weakrefs:
            class_mems.append("__weakref__")

        # In the future we may want a compatibility mode where we add
        # __dict__ and __weakref__
        gen.emit("LOAD_CONST", tuple(class_mems))
        gen.emit("STORE_NAME", "__slots__")

        slots_with_default = [
            name
            for name in class_mems_with_overrides
            if name in klass.members
            and isinstance(klass.members[name], Slot)
            and cast(
                Slot[Class], klass.members[name]
            ).is_typed_descriptor_with_default_value()
        ]

        for name in slots_with_default:
            gen.emit("LOAD_CONST", name)
            gen.emit("LOAD_NAME", name)
            gen.emit("DELETE_NAME", name)

        gen.emit("BUILD_MAP", len(slots_with_default))
        gen.emit("STORE_NAME", "__slots_with_default__")

        count = 0
        for name, value in klass.members.items():
            if not isinstance(value, Slot):
                continue

            if value.is_classvar:
                continue

            value_type = value.decl_type
            if value.decl_type is self.compiler.type_env.dynamic:
                if value.is_typed_descriptor_with_default_value():
                    value_type = self.compiler.type_env.object
                else:
                    continue

            gen.emit("LOAD_CONST", name)
            gen.emit("LOAD_CONST", value_type.type_descr)
            count += 1

        if count:
            gen.emit("BUILD_MAP", count)
            gen.emit("STORE_NAME", "__slot_types__")

    def visit_decorator(self, decorator: AST, class_def: ClassDef) -> None:
        d = self.get_type(decorator)
        if (
            isinstance(d, DataclassDecorator)
            and self.get_type(class_def) is not self.compiler.type_env.dynamic
        ):
            return

        super().visit_decorator(decorator, class_def)

    def emit_decorator_call(self, decorator: AST, class_def: ClassDef) -> None:
        self.get_type(decorator).emit_decorator_call(class_def, self)

    def emit_load_builtin(self, name: str) -> None:
        is_checked_dict = (
            name == "dict" and ModuleFlag.CHECKED_DICTS in self.cur_mod.flags
        )
        is_checked_list = (
            name == "list" and ModuleFlag.CHECKED_LISTS in self.cur_mod.flags
        )
        if is_checked_dict or is_checked_list:
            chkname = f"chk{name}"
            self.emit("LOAD_CONST", 0)
            self.emit("LOAD_CONST", (chkname,))
            self.emit(EAGER_IMPORT_NAME, "_static")
            self.emit("IMPORT_FROM", chkname)
            self.emit_rotate_stack(2)
            self.emit("POP_TOP")

    def visitModule(self, node: Module) -> None:
        if ModuleFlag.CHECKED_DICTS in self.cur_mod.flags:
            self.emit_load_builtin("dict")
            self.storeName("dict")

        if ModuleFlag.CHECKED_LISTS in self.cur_mod.flags:
            self.emit_load_builtin("list")
            self.storeName("list")

        super().visitModule(node)

    def emit_module_return(self, node: ast.Module) -> None:
        self.emit("LOAD_CONST", tuple(self.cur_mod.named_finals.keys()))
        self.emit("STORE_NAME", "__final_constants__")
        super().emit_module_return(node)

    def visitAssert(self, node: ast.Assert) -> None:
        super().visitAssert(node)
        # Only add casts when the assert is optimized out.
        if not self.optimization_lvl:
            return
        # Since we're narrowing types in asserts, we need to ensure we cast
        # all narrowed locals when asserts are optimized away.
        effect = self.get_node_data(node, NarrowingEffect)
        # As all of our effects store the final type, we can apply the effects on
        # an empty dictionary to get an overapproximation of what we need to cast.
        type_state = TypeState()
        effect_nodes: dict[str, ast.AST] = {}
        effect.apply(type_state, effect_nodes)
        for key, value in type_state.local_types.items():
            if value.klass.is_nominal_type:
                self.visit(effect_nodes[key])
                self.emit("CAST", value.klass.type_descr)
                self.emit("POP_TOP")
        for base, refinement_dict in type_state.refined_fields.items():
            for attr, (value, _, _) in refinement_dict.items():
                if value.klass.is_nominal_type:
                    key = f"{base}.{attr}"
                    self.visit(effect_nodes[key])
                    self.emit("CAST", value.klass.type_descr)
                    self.emit("POP_TOP")

    def visitAttribute(self, node: Attribute) -> None:
        self.set_pos(node)
        data = self.get_opt_node_data(node, UsedRefinementField)
        if data is not None and not data.is_source:
            self.emit("LOAD_FAST", data.name)
            return

        if isinstance(node.ctx, ast.Store) and data is not None and data.is_used:
            self.emit_dup()
            self.emit("STORE_FAST", data.name)

        self.get_type(node.value).emit_attr(node, self)

        if (
            isinstance(node.ctx, ast.Load)
            and data is not None
            and data.is_source
            and data.is_used
        ):
            self.emit_dup()
            self.emit("STORE_FAST", data.name)

    def visitAssignTarget(
        self, elt: expr, stmt: AST, value: expr | None = None
    ) -> None:
        if isinstance(elt, (ast.Tuple, ast.List)):
            self._visitUnpack(elt)
            if isinstance(value, ast.Tuple) and len(value.elts) == len(elt.elts):
                for target, inner_value in zip(elt.elts, value.elts):
                    self.visitAssignTarget(target, stmt, inner_value)
            else:
                for target in elt.elts:
                    self.visitAssignTarget(target, stmt, None)
        else:
            elt_type = self.get_type(elt).klass
            value_type = (
                self.compiler.type_env.dynamic
                if value is None
                else self.get_type(value).klass
            )
            elt_type.emit_type_check(value_type, self)
            self.visit(elt)

    def visitAssign(self, node: Assign) -> None:
        self.set_pos(node)
        self.visit(node.value)
        dups = len(node.targets) - 1
        for i in range(len(node.targets)):
            elt = node.targets[i]
            if i < dups:
                self.emit_dup()
            if isinstance(elt, ast.AST):
                self.visitAssignTarget(elt, node, node.value)

    def visitAnnAssign(self, node: ast.AnnAssign) -> None:
        self.set_pos(node)
        value = node.value
        if value:
            value_type = self.get_type(value)
            if (
                isinstance(value_type, DataclassField)
                and isinstance(self.tree, ast.ClassDef)
                and isinstance(self.get_type(self.tree), Dataclass)
            ):
                value_type.emit_field(node.target, self)
            else:
                self.visit(value)
                self.get_type(node.target).klass.emit_type_check(value_type.klass, self)
                self.visit(node.target)
        target = node.target
        if isinstance(target, ast.Name):
            self.emit_simple_ann_assign(node)
        elif isinstance(target, ast.Attribute):
            if not node.value:
                self.checkAnnExpr(target.value)
        elif isinstance(target, ast.Subscript):
            if not node.value:
                self.checkAnnExpr(target.value)
                self.checkAnnSubscr(target.slice)
        else:
            raise SystemError(
                f"invalid node type {type(node).__name__} for annotated assignment"
            )

        if not node.simple:
            self.checkAnnotation(node)

    def visitConstant(self, node: Constant) -> None:
        self.get_type(node).emit_constant(node, self)

    def visitDefault(self, node: expr) -> None:
        self.visit(node)
        node_type = self.get_type(node)
        if isinstance(node_type, CInstance):
            node_type.emit_box(self)

    def get_final_literal(self, node: AST) -> ast.Constant | None:
        return self.cur_mod.get_final_literal(node, self.scope)

    def visitName(self, node: Name) -> None:
        final_val = self.get_final_literal(node)
        if final_val is not None:
            # visit the constant directly
            return self.defaultVisit(final_val)
        self.get_type(node).emit_name(node, self)

    def emitAugAttribute(self, node: ast.AugAssign) -> None:
        target = node.target
        assert isinstance(target, ast.Attribute)
        self.visit(target.value)
        typ = self.get_type(target.value)
        self.emit_dup()
        typ.emit_load_attr(target, self)
        self.emitAugRHS(node)
        self.emit_rotate_stack(2)
        typ.emit_store_attr(target, self)

    def emitAugName(self, node: ast.AugAssign) -> None:
        target = node.target
        assert isinstance(target, ast.Name)
        typ = self.get_type(target)
        typ.emit_load_name(target, self)
        self.emitAugRHS(node)
        typ.emit_store_name(target, self)

    def emitAugSubscript(self, node: ast.AugAssign) -> None:
        target = node.target
        assert isinstance(target, ast.Subscript)
        self.visit(target.value)
        self.visit(target.slice)
        typ = self.get_type(target.value)
        self.emit_dup(2)
        typ.emit_load_subscr(target, self)
        self.emitAugRHS(node)
        self.emit_rotate_stack(3)
        typ.emit_store_subscr(target, self)

    def emitAugRHS(self, node: ast.AugAssign) -> None:
        self.get_type(node.target).emit_aug_rhs(node, self)

    def visitBinOp(self, node: BinOp) -> None:
        self.get_type(node).emit_binop(node, self)

    def visitUnaryOp(self, node: UnaryOp, type_ctx: Class | None = None) -> None:
        self.get_type(node).emit_unaryop(node, self)

    def visitCall(self, node: Call) -> None:
        self.strictPreVisitCall(node)
        self.get_type(node.func).emit_call(node, self)

    def visitSubscript(self, node: ast.Subscript, aug_flag: bool = False) -> None:
        # aug_flag is unused in static compiler; we have our own
        # emitAugSubscript that doesn't call visitSubscript
        self.get_type(node.value).emit_subscr(node, self)

    def _visitReturnValue(self, value: ast.AST, expected: Class) -> None:
        self.visit(value)
        expected.emit_type_check(self.get_type(value).klass, self)

    def visitReturn(self, node: ast.Return) -> None:
        self.checkReturn(node)
        function = self.get_func_container(self.tree)

        expected = function.return_type.resolved()
        if isinstance(self.tree, AsyncFunctionDef):
            assert isinstance(expected, AwaitableType)
            expected = expected.type_args[0]
        self.set_pos(node)
        value = node.value
        is_return_constant = isinstance(value, ast.Constant)
        opcode = "RETURN_VALUE"
        oparg = 0
        if value:
            if not is_return_constant:
                self._visitReturnValue(value, expected)
                self.unwind_setup_entries(preserve_tos=True)
            else:
                self.unwind_setup_entries(preserve_tos=False)
                self._visitReturnValue(value, expected)
            if isinstance(expected, CType):
                opcode = "RETURN_PRIMITIVE"
                oparg = expected.instance.as_oparg()
        else:
            self.unwind_setup_entries(preserve_tos=False)
            self.emit("LOAD_CONST", None)

        self.emit(opcode, oparg)
        self.nextBlock()

    def visitDictComp(self, node: DictComp) -> None:
        dict_type = self.get_type(node)
        if dict_type in (
            self.compiler.type_env.dict.instance,
            self.compiler.type_env.dict.exact_type().instance,
        ):
            return super().visitDictComp(node)
        klass = dict_type.klass

        assert (
            isinstance(klass, GenericClass)
            and klass.type_def is self.compiler.type_env.checked_dict
        ), dict_type
        self.compile_comprehension(
            node,
            sys.intern("<dictcomp>"),
            node.key,
            node.value,
            "BUILD_CHECKED_MAP",
            (dict_type.klass.type_descr, 0),
        )

    def compile_subgendict(
        self, node: ast.Dict, begin: int, end: int, dict_descr: TypeDescr
    ) -> None:
        n = end - begin
        for i in range(begin, end):
            k = node.keys[i]
            assert k is not None
            self.visit(k)
            self.visit(node.values[i])

        self.emit("BUILD_CHECKED_MAP", (dict_descr, n))

    def visitDict(self, node: ast.Dict) -> None:
        dict_type = self.get_type(node)
        if dict_type in (
            self.compiler.type_env.dict.instance,
            self.compiler.type_env.dict.exact_type().instance,
        ):
            return super().visitDict(node)
        klass = dict_type.klass

        assert (
            isinstance(klass, GenericClass)
            and klass.type_def is self.compiler.type_env.checked_dict
        ), dict_type

        self.set_pos(node)
        elements = 0
        is_unpacking = False
        built_final_dict = False

        # This is similar to the normal dict code generation, but instead of relying
        # upon an opcode for BUILD_MAP_UNPACK we invoke the update method on the
        # underlying dict type.  Therefore the first dict that we create becomes
        # the final dict.  This allows us to not introduce a new opcode, but we should
        # also be able to dispatch the INVOKE_METHOD rather efficiently.
        dict_descr = dict_type.klass.type_descr
        update_descr = dict_descr + ("update",)
        for i, (k, v) in enumerate(zip(node.keys, node.values)):
            is_unpacking = k is None
            if elements == 0xFFFF or (elements and is_unpacking):
                self.compile_subgendict(node, i - elements, i, dict_descr)
                built_final_dict = True
                elements = 0

            if is_unpacking:
                if not built_final_dict:
                    # {**foo, ...}, we need to generate the empty dict
                    self.emit("BUILD_CHECKED_MAP", (dict_descr, 0))
                    built_final_dict = True
                self.emit_dup()
                self.visit(v)

                self.emit_invoke_method(update_descr, 1)
                self.emit("POP_TOP")
            else:
                elements += 1

        if elements or not built_final_dict:
            if built_final_dict:
                self.emit_dup()
            self.compile_subgendict(
                node, len(node.keys) - elements, len(node.keys), dict_descr
            )
            if built_final_dict:
                self.emit_invoke_method(update_descr, 1)
                self.emit("POP_TOP")

    def compile_sub_checked_list(
        self, node: ast.List, begin: int, end: int, type_descr: TypeDescr
    ) -> None:
        n = end - begin
        for i in range(begin, end):
            elt = node.elts[i]
            assert not isinstance(elt, ast.Starred)
            self.visit(elt)

        self.emit("BUILD_CHECKED_LIST", (type_descr, n))

    def visitListComp(self, node: ast.ListComp) -> None:
        list_type = self.get_type(node)
        if list_type in (
            self.compiler.type_env.list.instance,
            self.compiler.type_env.list.exact_type().instance,
        ):
            return super().visitListComp(node)
        klass = list_type.klass

        assert (
            isinstance(klass, GenericClass)
            and klass.type_def is self.compiler.type_env.checked_list
        ), list_type
        self.compile_comprehension(
            node,
            sys.intern("<listcomp>"),
            node.elt,
            None,
            "BUILD_CHECKED_LIST",
            (list_type.klass.type_descr, 0),
        )

    def visitList(self, node: ast.List) -> None:
        list_type = self.get_type(node)
        if list_type in (
            self.compiler.type_env.list.instance,
            self.compiler.type_env.list.exact_type().instance,
        ):
            return super().visitList(node)
        klass = list_type.klass

        assert (
            isinstance(klass, GenericClass)
            and klass.type_def is self.compiler.type_env.checked_list
        ), list_type

        self.set_pos(node)
        list_descr = list_type.klass.type_descr
        extend_descr = (list_descr, "extend")
        built_final_list = False
        elements = 0

        for i, elt in enumerate(node.elts):
            if isinstance(elt, ast.Starred):
                if elements:
                    self.compile_sub_checked_list(node, i - elements, i, list_descr)
                    built_final_list = True
                    elements = 0
                if not built_final_list:
                    # We need to generate the empty list to extend in the case of [*foo, ...].
                    self.emit("BUILD_CHECKED_LIST", (list_descr, 0))
                    built_final_list = True
                self.emit_dup()
                self.emit_load_static_method(extend_descr)
                self.visit(elt.value)
                self.emit_invoke_method(extend_descr, 1)
                self.emit("POP_TOP")
            else:
                elements += 1

        if elements or not built_final_list:
            if built_final_list:
                self.emit_dup()
            self.compile_sub_checked_list(
                node, len(node.elts) - elements, len(node.elts), list_descr
            )
            if built_final_list:
                self.emit_invoke_method(extend_descr, 1)
                self.emit("POP_TOP")

    def visitContinue(self, node: ast.Continue) -> None:
        loop_node = self.get_opt_node_data(node, AST)
        if isinstance(loop_node, ast.For):
            # Since we support primitive int for loops with crange()
            # we have some special handling to deal with the "continue"
            # statement.
            iter_type = self.get_type(loop_node.iter)
            iter_type.emit_continue(node, self)
        else:
            super().visitContinue(node)

    def visitFor(self, node: ast.For) -> None:
        self.strictPreVisitFor(node)
        iter_type = self.get_type(node.iter)
        iter_type.emit_forloop(node, self)
        self.strictPostVisitFor(node)

    def emit_invoke_function(self, descr: TypeDescr, arg_count: int) -> None:
        # Emit a zero EXTENDED_ARG before so that we can optimize and insert the
        # arg count
        self.emit("EXTENDED_ARG", 0)
        self.emit("INVOKE_FUNCTION", (descr, arg_count))

    def emit_invoke_method(
        self, descr: TypeDescr, arg_count: int, is_classmethod: bool = False
    ) -> None:
        # Emit a zero EXTENDED_ARG before so that we can optimize and insert the
        # arg count
        self.emit("EXTENDED_ARG", 0)
        self.emit(
            "INVOKE_METHOD",
            (descr, arg_count, True) if is_classmethod else (descr, arg_count),
        )

    def emit_load_static_method(
        self, descr: TypeDescr, is_classmethod: bool = False
    ) -> None:
        # Emit a zero EXTENDED_ARG before so that we can optimize and insert the
        # arg count
        self.emit("EXTENDED_ARG", 0)
        self.emit(
            "LOAD_METHOD_STATIC",
            (descr, True) if is_classmethod else (descr,),
        )

    def defaultCall(self, node: object, name: str, *args: object) -> None:
        meth = getattr(self.parent_impl, name)
        return meth(self, node, *args)

    def defaultVisit(self, node: AST, *args: object) -> None:
        klass = node.__class__
        meth = self._default_cache.get(klass, None)
        if meth is None:
            className = klass.__name__
            meth = getattr(
                super(StaticCodeGenBase, StaticCodeGenBase),
                "visit" + className,
                StaticCodeGenBase.generic_visit,
            )
            self._default_cache[klass] = meth
        return meth(self, node, *args)

    def get_bool_const(self, node: AST) -> bool | None:
        if isinstance(node, ast.Constant):
            return bool(node.value)

        kb = self.get_opt_node_data(node, KnownBoolean)
        if kb is not None:
            return True if kb == KnownBoolean.TRUE else False

    def visitIf(self, node: ast.If) -> None:
        self.get_type(node.test)

        test_const = self.get_bool_const(node.test)

        end = self.newBlock("if_end")
        orelse = None
        if node.orelse:
            orelse = self.newBlock("if_else")

        self.compileJumpIf(node.test, orelse or end, False)
        if test_const is not False:
            self.visitStatements(node.body)

        if node.orelse:
            self.emit_jump_forward_noline(end)
            self.nextBlock(orelse)
            if test_const is not True:
                self.visitStatements(node.orelse)

        self.nextBlock(end)

    def compileJumpIf(self, test: AST, next: Block, is_if_true: bool) -> None:
        if isinstance(test, ast.UnaryOp):
            if isinstance(test.op, ast.Not):
                # Compile to remove not operation
                return self.compileJumpIf(test.operand, next, not is_if_true)
        elif isinstance(test, ast.BoolOp):
            is_or = isinstance(test.op, ast.Or)
            skip_jump = next
            if is_if_true != is_or:
                skip_jump = self.newBlock()

            for node in test.values[:-1]:
                self.get_type(node).emit_jumpif(node, skip_jump, is_or, self)
                self.nextBlock()

            self.get_type(test.values[-1]).emit_jumpif(
                test.values[-1], next, is_if_true, self
            )
            self.nextBlock()

            if skip_jump is not next:
                self.nextBlock(skip_jump)
            return
        elif isinstance(test, ast.IfExp):
            end = self.newBlock("end")
            orelse = self.newBlock("orelse")
            # Jump directly to orelse if test matches
            self.get_type(test.test).emit_jumpif(test.test, orelse, False, self)
            # Jump directly to target if test is true and body is matches
            self.get_type(test.body).emit_jumpif(test.body, next, is_if_true, self)
            self.emit_jump_forward_noline(end)
            # Jump directly to target if test is true and orelse matches
            self.nextBlock(orelse)
            self.get_type(test.orelse).emit_jumpif(test.orelse, next, is_if_true, self)
            self.nextBlock(end)
            return

        self.get_type(test).emit_jumpif(test, next, is_if_true, self)
        self.nextBlock()

    def _calculate_idx(
        self, arg_name: str, non_cellvar_pos: int, cellvars: IndexedSet
    ) -> int:
        if sys.version_info >= (3, 12):
            return non_cellvar_pos
        try:
            offset = cellvars.index(arg_name)
        except ValueError:
            return non_cellvar_pos
        else:
            # the negative sign indicates to the runtime/JIT that this is a cellvar
            return -(offset + 1)

    def perf_warning(self, msg: str, node: AST) -> None:
        return self.compiler.error_sink.perf_warning(msg, self.graph.filename, node)

    def storePatternName(self, name: str, pc: PatternContext) -> None:
        pc = cast(StaticPatternContext, pc)
        pc.type_dict[name].emit_type_check(self.compiler.type_env.DYNAMIC.klass, self)
        self.storeName(name)

    def _pattern_helper_store_name(
        self, name: str | None, pc: PatternContext, loc: ast.AST
    ) -> None:
        super()._pattern_helper_store_name(name, pc, loc)
        pc = cast(StaticPatternContext, pc)
        if name is not None:
            pc.type_dict[name] = self.get_type(loc).klass

    def emit_convert_compare_arg(self, op: AST, optype: Value, ltype: Value) -> None:
        """Convert `ltype` to be compatible with `optype`."""
        # Helper function for visitCompare; checks if ltype needs to be boxed.
        optype.emit_convert(ltype, self)
        if (
            isinstance(ltype, CInstance)
            and isinstance(optype, DynamicInstance)
            and isinstance(op, ast.In)
        ):
            # We don't support static containers yet, so for something
            # like `x: int64 in y: list` we set the optype to dynamic
            # in static.types.CIntInstance.bind_compare, and we need to
            # box `x` here before the COMPARE opcode is emitted.
            ltype.emit_box(self)


class Static310CodeGenerator(StaticCodeGenBase, CinderCodeGenerator310):
    flow_graph = PyFlowGraphStatic310
    parent_impl = StrictCodeGenerator310

    def visitBoolOp(self, node: BoolOp) -> None:
        end = self.newBlock()
        for child in node.values[:-1]:
            self.get_type(child).emit_jumpif_pop(
                child, end, type(node.op) is ast.Or, self
            )
            self.nextBlock()
        self.visit(node.values[-1])
        self.nextBlock(end)

    def visitCompare(self, node: Compare) -> None:
        self.set_pos(node)
        self.visit(node.left)
        cleanup = self.newBlock("cleanup")
        left = node.left
        for op, code in zip(node.ops[:-1], node.comparators[:-1]):
            optype = self.get_type(op)
            ltype = self.get_type(left)
            if ltype != optype:
                optype.emit_convert(ltype, self)
            self.emitChainedCompareStep(op, code, cleanup)
            left = code
        # now do the last comparison
        if node.ops:
            op = node.ops[-1]
            optype = self.get_type(op)
            ltype = self.get_type(left)
            if ltype != optype:
                self.emit_convert_compare_arg(op, optype, ltype)
            code = node.comparators[-1]
            self.visit(code)
            rtype = self.get_type(code)
            if rtype != optype:
                optype.emit_convert(rtype, self)
            optype.emit_compare(op, self)
        if len(node.ops) > 1:
            end = self.newBlock("end")
            self.emit_jump_forward(end)
            self.nextBlock(cleanup)
            self.emit_rotate_stack(2)
            self.emit("POP_TOP")
            self.nextBlock(end)

    def emitChainedCompareStep(
        self, op: cmpop, value: AST, cleanup: Block, always_pop: bool = False
    ) -> None:
        optype = self.get_type(op)
        self.visit(value)
        rtype = self.get_type(value)
        if rtype != optype:
            optype.emit_convert(rtype, self)
        self.emit_dup()
        self.emit_rotate_stack(3)
        optype.emit_compare(op, self)
        method = optype.emit_jumpif_only if always_pop else optype.emit_jumpif_pop_only
        method(cleanup, False, self)
        self.nextBlock(label="compare_or_cleanup")


class Static312CodeGenerator(StaticCodeGenBase, CinderCodeGenerator312):
    flow_graph = PyFlowGraph312
    parent_impl = StrictCodeGenerator312

    def visitBoolOp(self, node: ast.BoolOp) -> None:
        is_if_true = type(node.op) is not ast.And
        end = self.newBlock()
        for value in node.values[:-1]:
            self.visit(value)
            self.emit("COPY", 1)
            self.get_type(value).emit_jumpif_only(end, is_if_true, self)
            self.nextBlock()
            self.emit("POP_TOP")
        self.visit(node.values[-1])
        self.nextBlock(end)

    def visitCompare(self, node: Compare) -> None:
        self.set_pos(node)
        left = node.left
        self.visit(left)
        ltype = self.get_type(left)

        if len(node.comparators) == 1:
            op = node.ops[0]
            optype = self.get_type(op)
            if ltype != optype:
                self.emit_convert_compare_arg(op, optype, ltype)

            self.visit(node.comparators[0])
            rtype = self.get_type(node.comparators[0])
            if rtype != optype:
                optype.emit_convert(rtype, self)

            optype.emit_compare(op, self)
            return

        cleanup = self.newBlock("cleanup")
        for i in range(len(node.comparators) - 1):
            self.set_pos(node)
            op = node.ops[i]
            optype = self.get_type(op)
            if ltype != optype:
                self.emit_convert_compare_arg(op, optype, ltype)

            self.emitChainedCompareStep(
                node.ops[i], node.comparators[i], cleanup, False
            )
            left = node.comparators[i]
            self.emit("COPY", 1)
            optype = self.get_type(node.ops[i])
            optype.emit_jumpif_only(cleanup, False, self)
            self.nextBlock(label="compare_or_cleanup")
            self.emit("POP_TOP")

        op = node.ops[-1]
        optype = self.get_type(op)
        ltype = self.get_type(left)
        if ltype != optype:
            optype.emit_convert(ltype, self)

        self.visit(node.comparators[-1])

        rtype = self.get_type(node.comparators[-1])
        optype.emit_compare(op, self)
        end = self.newBlock("end")
        self.emit_jump_forward(end)

        self.nextBlock(cleanup)
        self.emit("SWAP", 2)
        self.emit("POP_TOP")
        self.nextBlock(end)

    def emitChainedCompareStep(
        self,
        op: ast.cmpop,
        value: ast.expr,
        cleanup: Block,
        always_pop: bool = False,
    ) -> None:
        optype = self.get_type(op)
        self.visit(value)
        rtype = self.get_type(value)
        if rtype != optype:
            optype.emit_convert(rtype, self)
        self.emit("SWAP", 2)
        self.emit("COPY", 2)
        optype = self.get_type(op)
        optype.emit_compare(op, self)


class Static314CodeGenerator(Static312CodeGenerator, CinderCodeGenerator314):
    flow_graph = PyFlowGraph314
    parent_impl = StrictCodeGenerator314

    def emit_invoke_function(self, descr: TypeDescr, arg_count: int) -> None:
        self.emit("INVOKE_FUNCTION", (descr, arg_count))

    def emit_invoke_method(
        self, descr: TypeDescr, arg_count: int, is_classmethod: bool = False
    ) -> None:
        self.emit(
            "INVOKE_METHOD",
            (descr, arg_count, True) if is_classmethod else (descr, arg_count),
        )

    def emit_load_static_method(
        self, descr: TypeDescr, is_classmethod: bool = False
    ) -> None:
        self.emit(
            "LOAD_METHOD_STATIC",
            (descr, True) if is_classmethod else (descr,),
        )


class Static315CodeGenerator(Static314CodeGenerator, CinderCodeGenerator315):
    flow_graph = PyFlowGraph315


if sys.version_info >= (3, 15):
    StaticCodeGenerator = Static315CodeGenerator
elif sys.version_info >= (3, 14):
    StaticCodeGenerator = Static314CodeGenerator
elif sys.version_info >= (3, 12):
    StaticCodeGenerator = Static312CodeGenerator
else:
    StaticCodeGenerator = Static310CodeGenerator



# Copyright (c) Meta Platforms, Inc. and affiliates.

# pyre-strict

from __future__ import annotations

import ast
from ast import AST, Attribute, Call, Compare, Constant, Expr, Name, Return
from collections.abc import Sequence
from typing import TYPE_CHECKING

from ..errors import TypedSyntaxError
from ..symbols import SymbolVisitor
from .effects import NarrowingEffect
from .pyrefly_info import PyreflyTypeInfo
from .type_binder import (
    PRESERVE_REFINED_FIELDS,
    PreserveRefinedFields,
    TerminalKind,
    TypeBinder,
)
from .types import (
    # CInstance,
    # Class,
    # CType,
    Dataclass,
    DataclassField,
    ModuleInstance,
    TypeDescr,
)

if TYPE_CHECKING:
    from .compiler import Compiler


class PyreflyTypeBinder(TypeBinder):
    """TypeBinder that uses pyrefly type inference to set types on expression nodes.

    For each expression node, looks up the pyrefly-inferred type,
    resolves it via the module table, and sets it on the node.
    """

    def __init__(
        self,
        symbols: SymbolVisitor,
        filename: str,
        compiler: Compiler,
        module_name: str,
        optimize: int,
        enable_patching: bool = False,
        type_info: PyreflyTypeInfo | None = None,
    ) -> None:
        super().__init__(
            symbols, filename, compiler, module_name, optimize, enable_patching
        )
        assert type_info is not None
        self._type_info = type_info

    def visit(self, node: AST, *args: object) -> NarrowingEffect | None:
        if isinstance(node, ast.expr) and self._type_info is not None:
            ret = super().generic_visit(node, *args)
            # pyre-fixme[16]: Optional type has no attribute `lookup`.
            declared_type = self._type_info.lookup(node, self.modules, self.type_env)
            if declared_type is None:
                declared_type = self.type_env.dynamic.instance
                if isinstance(node, (ast.List, ast.ListComp)):
                    declared_type = self.type_env.list.instance
                elif isinstance(node, (ast.Dict, ast.DictComp)):
                    declared_type = self.type_env.dict.instance
            self.set_type(node, declared_type)

            if isinstance(node, Compare):
                for op in node.ops:
                    self.set_type(op, self.type_env.DYNAMIC)

            # Name: set PreserveRefinedFields (always), declare locals for
            # Store context, and set TypeDescr for module-level names so
            # that bind_call can emit direct invocations.  When pyrefly
            # doesn't resolve the type, fall back to the module table so
            # CinderX-specific types (e.g. ModuleInstance) are preserved.
            elif isinstance(node, Name):
                self.set_node_data(node, PreserveRefinedFields, PRESERVE_REFINED_FIELDS)
                if isinstance(node.ctx, ast.Store):
                    try:
                        self.declare_local(node.id, declared_type)
                    except TypedSyntaxError:
                        pass  # already declared, just update the type
                mod_typ, descr = self.module.resolve_name_with_descr(
                    node.id, self.context_qualname
                )
                if descr is not None:
                    self.set_node_data(node, TypeDescr, descr)
                if (
                    mod_typ is not None
                    and declared_type is self.type_env.dynamic.instance
                ):
                    self.set_type(node, mod_typ)

            # Attribute: when the base is a ModuleInstance, set TypeDescr
            # for direct access and call bind_attr to resolve from the
            # CinderX module table — this ensures CinderX-specific types
            # (e.g. DataclassFieldFunction, DataclassDecorator) are used.
            # Set PreserveRefinedFields when the attribute is refinable.
            elif isinstance(node, Attribute):
                base = self.get_type(node.value)
                if isinstance(base, ModuleInstance):
                    self.set_node_data(
                        node, TypeDescr, ((base.module_name,), node.attr)
                    )
                    # Always call bind_attr for module attributes so that
                    # CinderX-specific types (e.g. DataclassFieldFunction,
                    # DataclassDecorator) are used regardless of what
                    # pyrefly resolved.
                    base.bind_attr(node, self, None)
                if self.is_refinable(node):
                    self.set_node_data(
                        node, PreserveRefinedFields, PRESERVE_REFINED_FIELDS
                    )

            # Call: invoke bind_call on the func's type to populate
            # ArgMapping (and ClassCallInfo for class instantiation).
            elif isinstance(node, Call):
                self.get_type(node.func).bind_call(node, self, None)
                # When pyrefly compiles modules that aren't normally
                # statically compiled, dataclasses.field() may resolve
                # to DataclassField via the CinderX module table.
                # In non-Dataclass classes (e.g. decorated with
                # @deprecated on top of @dataclass), this causes type
                # errors in visitAnnAssign.  Reset to dynamic and mark
                # the class for non-static compilation in that case.
                if isinstance(self.get_type(node), DataclassField) and not (
                    isinstance(self.scope, ast.ClassDef)
                    and isinstance(self.get_type(self.scope), Dataclass)
                ):
                    self.set_type(node, declared_type)
                    if isinstance(self.scope, ast.ClassDef):
                        self.module.compile_non_static.add(self.scope)
        else:
            return super().visit(node, *args)

        return ret

    def visit_check_terminal(self, nodes: Sequence[ast.stmt]) -> TerminalKind:
        # Treat a body consisting of just `...` (Ellipsis) as a terminal
        # statement, so that stub functions like Protocol methods don't
        # trigger "can implicitly return None" errors.
        if (
            len(nodes) == 1
            and isinstance(nodes[0], Expr)
            and isinstance(nodes[0].value, Constant)
            and nodes[0].value.value is ...
        ):
            self.visit(nodes[0])
            self.set_terminal_kind(nodes[0], TerminalKind.RaiseOrReturn)
            return TerminalKind.RaiseOrReturn
        return super().visit_check_terminal(nodes)

    def visitReturn(self, node: Return) -> None:
        self.set_terminal_kind(node, TerminalKind.RaiseOrReturn)
        if node.value is not None:
            self.visit(node.value)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# pyre-strict

from _cinderx import StaticTypeError
from _static import (
    __build_cinder_class__,
    _clear_dlopen_cache,
    _clear_dlsym_cache,
    _property_missing_fget,
    _property_missing_fset,
    _sizeof_dlopen_cache,
    _sizeof_dlsym_cache,
    chkdict,
    chklist,
    FAST_LEN_ARRAY,
    FAST_LEN_DICT,
    FAST_LEN_INEXACT,
    FAST_LEN_LIST,
    FAST_LEN_SET,
    FAST_LEN_STR,
    FAST_LEN_TUPLE,
    init_subclass,
    install_sp_audit_hook,
    is_static_callable,
    is_static_module,
    is_type_static,
    lookup_native_symbol,
    make_context_decorator_wrapper,
    make_recreate_cm,
    PRIM_OP_ADD_DBL,
    PRIM_OP_ADD_INT,
    PRIM_OP_AND_INT,
    PRIM_OP_DIV_DBL,
    PRIM_OP_DIV_INT,
    PRIM_OP_DIV_UN_INT,
    PRIM_OP_EQ_DBL,
    PRIM_OP_EQ_INT,
    PRIM_OP_GE_DBL,
    PRIM_OP_GE_INT,
    PRIM_OP_GE_UN_INT,
    PRIM_OP_GT_DBL,
    PRIM_OP_GT_INT,
    PRIM_OP_GT_UN_INT,
    PRIM_OP_INV_INT,
    PRIM_OP_LE_DBL,
    PRIM_OP_LE_INT,
    PRIM_OP_LE_UN_INT,
    PRIM_OP_LSHIFT_INT,
    PRIM_OP_LT_DBL,
    PRIM_OP_LT_INT,
    PRIM_OP_LT_UN_INT,
    PRIM_OP_MOD_DBL,
    PRIM_OP_MOD_INT,
    PRIM_OP_MOD_UN_INT,
    PRIM_OP_MUL_DBL,
    PRIM_OP_MUL_INT,
    PRIM_OP_NE_DBL,
    PRIM_OP_NE_INT,
    PRIM_OP_NEG_DBL,
    PRIM_OP_NEG_INT,
    PRIM_OP_NOT_INT,
    PRIM_OP_OR_INT,
    PRIM_OP_POW_DBL,
    PRIM_OP_POW_INT,
    PRIM_OP_POW_UN_INT,
    PRIM_OP_RSHIFT_INT,
    PRIM_OP_RSHIFT_UN_INT,
    PRIM_OP_SUB_DBL,
    PRIM_OP_SUB_INT,
    PRIM_OP_XOR_INT,
    rand,
    RAND_MAX,
    resolve_primitive_descr,
    SEQ_ARRAY_INT64,
    SEQ_CHECKED_LIST,
    SEQ_LIST,
    SEQ_LIST_INEXACT,
    SEQ_REPEAT_INEXACT_NUM,
    SEQ_REPEAT_INEXACT_SEQ,
    SEQ_REPEAT_PRIMITIVE_NUM,
    SEQ_REPEAT_REVERSED,
    SEQ_SUBSCR_UNCHECKED,
    SEQ_TUPLE,
    set_type_code,
    set_type_final,
    set_type_static,
    set_type_static_final,
    staticarray,
    TYPED_ARRAY,
    TYPED_BOOL,
    TYPED_CHAR,
    TYPED_DOUBLE,
    TYPED_INT16,
    TYPED_INT32,
    TYPED_INT64,
    TYPED_INT8,
    TYPED_INT_16BIT,
    TYPED_INT_32BIT,
    TYPED_INT_64BIT,
    TYPED_OBJECT,
    TYPED_SINGLE,
    TYPED_UINT16,
    TYPED_UINT32,
    TYPED_UINT64,
    TYPED_UINT8,
)



# Copyright (c) Meta Platforms, Inc. and affiliates.

# pyre-strict

from __future__ import annotations

from __static__ import rand, RAND_MAX

import ast
import builtins
from ast import AST
from collections import deque
from types import CodeType
from typing import Any, cast, TYPE_CHECKING

from .. import consts
from ..errors import ErrorSink
from ..optimizer import AstOptimizer
from ..pycodegen import find_futures
from ..symbols import SymbolVisitor
from .declaration_visitor import DeclarationVisitor
from .module_table import IntrinsicModuleTable, ModuleTable
from .type_binder import TypeBinder
from .types import (
    BoxFunction,
    CastFunction,
    Class,
    CRangeFunction,
    ExtremumFunction,
    IdentityDecorator,
    IsInstanceFunction,
    IsSubclassFunction,
    LenFunction,
    NumClass,
    Object,
    ProdAssertFunction,
    reflect_builtin_function,
    RevealTypeFunction,
    SortedFunction,
    TypeEnvironment,
    TypeName,
    UnboxFunction,
    Value,
)


if TYPE_CHECKING:
    from . import Static310CodeGenerator, StaticCodeGenBase

try:
    import xxclassloader
except ImportError:
    xxclassloader = None


class StrictBuiltins(Object[Class]):
    def __init__(self, builtins: dict[str, Value], type_env: TypeEnvironment) -> None:
        super().__init__(type_env.dict)
        self.builtins = builtins

    def bind_subscr(
        self,
        node: ast.Subscript,
        type: Value,
        visitor: TypeBinder,
        type_ctx: Class | None = None,
    ) -> None:
        slice = node.slice
        type = visitor.type_env.DYNAMIC
        if isinstance(slice, ast.Constant):
            svalue = slice.value
            if isinstance(svalue, str):
                builtin = self.builtins.get(svalue)
                if builtin is not None:
                    type = builtin

        visitor.set_type(node, type)


class Compiler:
    def __init__(
        self,
        code_generator: type[StaticCodeGenBase],
        error_sink: ErrorSink | None = None,
    ) -> None:
        self.modules: dict[str, ModuleTable] = {}
        self.ast_cache: dict[
            str | bytes | ast.Module | ast.Expression | ast.Interactive, ast.Module
        ] = {}
        self.code_generator = code_generator
        self.error_sink: ErrorSink = error_sink or ErrorSink()
        self.type_env: TypeEnvironment = TypeEnvironment()
        rand_max = NumClass(
            TypeName("builtins", "int"),
            self.type_env,
            pytype=int,
            # pyre-fixme[16]: Module `__static__` has no attribute `RAND_MAX`.
            literal_value=RAND_MAX,
            is_final=True,
        )
        builtins_children = {
            "ArithmeticError": self.type_env.arithmetic_error,
            "AssertionError": self.type_env.assertion_error,
            "AttributeError": self.type_env.attribute_error,
            "BaseException": self.type_env.base_exception.exact_type(),
            "BlockingIOError": self.type_env.blocking_io_error,
            "BrokenPipeError": self.type_env.broken_pipe_error,
            "BufferError": self.type_env.buffer_error,
            "BytesWarning": self.type_env.bytes_warning,
            "ChildProcessError": self.type_env.child_process_error,
            "ConnectionAbortedError": self.type_env.connection_aborted_error,
            "ConnectionError": self.type_env.connection_error,
            "ConnectionRefusedError": self.type_env.connection_refused_error,
            "ConnectionResetError": self.type_env.connection_reset_error,
            "DeprecationWarning": self.type_env.deprecation_warning,
            "EncodingWarning": self.type_env.encoding_warning,
            "EOFError": self.type_env.eof_error,
            "Ellipsis": self.type_env.DYNAMIC,
            "EnvironmentError": self.type_env.environment_error,
            "Exception": self.type_env.exception.exact_type(),
            "FileExistsError": self.type_env.file_exists_error,
            "FileNotFoundError": self.type_env.file_not_found_error,
            "FloatingPointError": self.type_env.floating_point_error,
            "FutureWarning": self.type_env.future_warning,
            "GeneratorExit": self.type_env.generator_exit,
            "IOError": self.type_env.io_error,
            "ImportError": self.type_env.import_error,
            "ImportWarning": self.type_env.import_warning,
            "IndentationError": self.type_env.indentation_error,
            "IndexError": self.type_env.index_error,
            "InterruptedError": self.type_env.interrupted_error,
            "IsADirectoryError": self.type_env.is_a_directory_error,
            "KeyError": self.type_env.key_error,
            "KeyboardInterrupt": self.type_env.keyboard_interrupt,
            "LookupError": self.type_env.lookup_error,
            "MemoryError": self.type_env.memory_error,
            "ModuleNotFoundError": self.type_env.module_not_found_error,
            "NameError": self.type_env.name_error,
            "None": self.type_env.none.instance,
            "NotADirectoryError": self.type_env.not_a_directory_error,
            "NotImplemented": self.type_env.not_implemented,
            "NotImplementedError": self.type_env.not_implemented_error,
            "OSError": self.type_env.os_error,
            "OverflowError": self.type_env.overflow_error,
            "PendingDeprecationWarning": self.type_env.pending_deprecation_warning,
            "PermissionError": self.type_env.permission_error,
            "ProcessLookupError": self.type_env.process_lookup_error,
            "RecursionError": self.type_env.recursion_error,
            "ReferenceError": self.type_env.reference_error,
            "ResourceWarning": self.type_env.resource_warning,
            "RuntimeError": self.type_env.runtime_error,
            "RuntimeWarning": self.type_env.runtime_warning,
            "StopAsyncIteration": self.type_env.stop_async_iteration,
            "StopIteration": self.type_env.stop_iteration,
            "SyntaxError": self.type_env.syntax_error,
            "SyntaxWarning": self.type_env.syntax_warning,
            "SystemError": self.type_env.system_error,
            "SystemExit": self.type_env.system_exit,
            "TabError": self.type_env.tab_error,
            "TimeoutError": self.type_env.timeout_error,
            "TypeError": self.type_env.type_error,
            "UnboundLocalError": self.type_env.unbound_local_error,
            "UnicodeDecodeError": self.type_env.unicode_decode_error,
            "UnicodeEncodeError": self.type_env.unicode_encode_error,
            "UnicodeError": self.type_env.unicode_error,
            "UnicodeTranslateError": self.type_env.unicode_translate_error,
            "UnicodeWarning": self.type_env.unicode_warning,
            "UserWarning": self.type_env.user_warning,
            "ValueError": self.type_env.value_error,
            "Warning": self.type_env.warning,
            "ZeroDivisionError": self.type_env.zero_division_error,
            "__name__": self.type_env.str.instance,
            "__file__": self.type_env.str.instance,
            "bool": self.type_env.bool.exact_type(),
            "bytes": self.type_env.bytes.exact_type(),
            "bytearray": self.type_env.DYNAMIC,
            "classmethod": self.type_env.class_method,
            "complex": self.type_env.complex.exact_type(),
            "delattr": self.type_env.DYNAMIC,
            "dict": self.type_env.dict.exact_type(),
            "enumerate": self.type_env.DYNAMIC,
            "exit": self.type_env.DYNAMIC,
            "float": self.type_env.float.exact_type(),
            "frozenset": self.type_env.frozenset.exact_type(),
            "getattr": self.type_env.DYNAMIC,
            "globals": self.type_env.DYNAMIC,
            "hasattr": self.type_env.DYNAMIC,
            "memoryview": self.type_env.DYNAMIC,
            "int": self.type_env.int.exact_type(),
            "isinstance": IsInstanceFunction(self.type_env),
            "issubclass": IsSubclassFunction(self.type_env),
            "len": LenFunction(self.type_env.function, boxed=True),
            "list": self.type_env.list.exact_type(),
            "locals": self.type_env.DYNAMIC,
            "max": ExtremumFunction(self.type_env.function, is_min=False),
            "min": ExtremumFunction(self.type_env.function, is_min=True),
            "object": self.type_env.object.exact_type(),
            "property": self.type_env.property.exact_type(),
            "quit": self.type_env.DYNAMIC,
            "range": self.type_env.range,
            "reveal_type": RevealTypeFunction(self.type_env),
            "set": self.type_env.set.exact_type(),
            "setattr": self.type_env.DYNAMIC,
            "slice": self.type_env.DYNAMIC,
            "sorted": SortedFunction(self.type_env.function),
            "staticmethod": self.type_env.static_method,
            "str": self.type_env.str.exact_type(),
            "super": self.type_env.super,
            "tuple": self.type_env.tuple.exact_type(),
            "type": self.type_env.type.exact_type(),
            "<mutable>": IdentityDecorator(
                TypeName("__strict__", "<mutable>"), self.type_env
            ),
            "filter": self.type_env.DYNAMIC,
            "map": self.type_env.DYNAMIC,
            "reversed": self.type_env.DYNAMIC,
            "zip": self.type_env.DYNAMIC,
            # pyre-ignore[6]: Pyre can't know this callable is a BuiltinFunctionType
            "abs": reflect_builtin_function(abs, None, self.type_env),
            # pyre-ignore[6]: Pyre can't know this callable is a BuiltinFunctionType
            "all": reflect_builtin_function(all, None, self.type_env),
            # pyre-ignore[6]: Pyre can't know this callable is a BuiltinFunctionType
            "any": reflect_builtin_function(any, None, self.type_env),
            "anext": reflect_builtin_function(anext, None, self.type_env),
            # pyre-ignore[6]: Pyre can't know this callable is a BuiltinFunctionType
            "ascii": reflect_builtin_function(ascii, None, self.type_env),
            # pyre-ignore[6]: Pyre can't know this callable is a BuiltinFunctionType
            "bin": reflect_builtin_function(bin, None, self.type_env),
            # pyre-ignore[6]: Pyre can't know this callable is a BuiltinFunctionType
            "breakpoint": reflect_builtin_function(breakpoint, None, self.type_env),
            # pyre-ignore[6]: Pyre can't know this callable is a BuiltinFunctionType
            "callable": reflect_builtin_function(callable, None, self.type_env),
            # pyre-ignore[6]: Pyre can't know this callable is a BuiltinFunctionType
            "chr": reflect_builtin_function(chr, None, self.type_env),
            "compile": reflect_builtin_function(compile, None, self.type_env),
            # pyre-ignore[6]: Pyre can't know this callable is a BuiltinFunctionType
            "dir": reflect_builtin_function(dir, None, self.type_env),
            "divmod": reflect_builtin_function(divmod, None, self.type_env),
            # pyre-ignore[6]: Pyre can't know this callable is a BuiltinFunctionType
            "eval": reflect_builtin_function(eval, None, self.type_env),
            # pyre-ignore[6]: Pyre can't know this callable is a BuiltinFunctionType
            "exec": reflect_builtin_function(exec, None, self.type_env),
            # pyre-ignore[6]: Pyre can't know this callable is a BuiltinFunctionType
            "format": reflect_builtin_function(format, None, self.type_env),
            # pyre-ignore[6]: Pyre can't know this callable is a BuiltinFunctionType
            "hash": reflect_builtin_function(hash, None, self.type_env),
            # pyre-ignore[6]: Pyre can't know this callable is a BuiltinFunctionType
            "hex": reflect_builtin_function(hex, None, self.type_env),
            # pyre-ignore[6]: Pyre can't know this callable is a BuiltinFunctionType
            "id": reflect_builtin_function(id, None, self.type_env),
            # pyre-ignore[6]: Pyre can't know this callable is a BuiltinFunctionType
            "input": reflect_builtin_function(input, None, self.type_env),
            "iter": reflect_builtin_function(iter, None, self.type_env),
            "next": reflect_builtin_function(next, None, self.type_env),
            # pyre-ignore[6]: Pyre can't know this callable is a BuiltinFunctionType
            "oct": reflect_builtin_function(oct, None, self.type_env),
            "open": reflect_builtin_function(open, None, self.type_env),
            # pyre-ignore[6]: Pyre can't know this callable is a BuiltinFunctionType
            "ord": reflect_builtin_function(ord, None, self.type_env),
            "pow": reflect_builtin_function(pow, None, self.type_env),
            "print": reflect_builtin_function(print, None, self.type_env),
            # pyre-ignore[6]: Pyre can't know this callable is a BuiltinFunctionType
            "repr": reflect_builtin_function(repr, None, self.type_env),
            "round": reflect_builtin_function(round, None, self.type_env),
            "sum": reflect_builtin_function(sum, None, self.type_env),
            "vars": reflect_builtin_function(vars, None, self.type_env),
            # FIXME: This is IG-specific. Add a way to customize the set of builtins
            # while initializing the loader.
            # List from https://www.internalfb.com/code/fbsource/[5f9467ab16be43bf56aa9d7f18580a14bff36de7]/fbcode/instagram-server/distillery/bootstrap/dev.py?lines=49
            "slog": self.type_env.DYNAMIC,
            "slogo": self.type_env.DYNAMIC,
            "slogb": self.type_env.DYNAMIC,
            "slogg": self.type_env.DYNAMIC,
            "slogv": self.type_env.DYNAMIC,
            "slogp": self.type_env.DYNAMIC,
            "slogy": self.type_env.DYNAMIC,
            "slogr": self.type_env.DYNAMIC,
            "slogdor": self.type_env.DYNAMIC,
            "sloga": self.type_env.DYNAMIC,
            "slogf": self.type_env.DYNAMIC,
            "slogrq": self.type_env.DYNAMIC,
            "slog0": self.type_env.DYNAMIC,
        }
        if import_cycle_error := self.type_env.import_cycle_error:
            builtins_children["ImportCycleError"] = import_cycle_error

        strict_builtins = StrictBuiltins(builtins_children, self.type_env)
        typing_children = {
            "Annotated": self.type_env.annotated,
            "Any": self.type_env.dynamic,
            "Awaitable": self.type_env.dynamic,
            "Callable": self.type_env.dynamic,
            "Coroutine": self.type_env.dynamic,
            "ClassVar": self.type_env.classvar,
            "Collection": self.type_env.dynamic,
            # TODO: Need typed members for dict
            "Dict": self.type_env.dict,
            "List": self.type_env.list,
            "Final": self.type_env.final,
            "final": self.type_env.final_method,
            "Generic": self.type_env.dynamic,
            "Iterable": self.type_env.dynamic,
            "Iterator": self.type_env.dynamic,
            "Literal": self.type_env.literal,
            "NamedTuple": self.type_env.named_tuple,
            "Protocol": self.type_env.protocol,
            "Optional": self.type_env.optional,
            "overload": self.type_env.overload,
            "Union": self.type_env.union,
            "Tuple": self.type_env.tuple,
            "Type": self.type_env.dynamic,
            "TypedDict": self.type_env.typed_dict,
            "TypeVar": self.type_env.dynamic,
            "TYPE_CHECKING": self.type_env.bool.instance,
        }
        typing_extensions_children: dict[str, Value] = {
            "Annotated": self.type_env.annotated,
            "NamedTuple": self.type_env.named_tuple,
            "Protocol": self.type_env.protocol,
            "TypedDict": self.type_env.typed_dict,
        }
        strict_modules_children: dict[str, Value] = {
            "mutable": IdentityDecorator(
                TypeName("__strict__", "mutable"), self.type_env
            ),
            "strict_slots": IdentityDecorator(
                TypeName("__strict__", "strict_slots"), self.type_env
            ),
            "loose_slots": IdentityDecorator(
                TypeName("__strict__", "loose_slots"), self.type_env
            ),
            "freeze_type": IdentityDecorator(
                TypeName("__strict__", "freeze_type"), self.type_env
            ),
        }

        builtins_children["<builtins>"] = strict_builtins
        self.modules["strict_modules"] = self.modules["__strict__"] = ModuleTable(
            "strict_modules", "<strict-modules>", self, strict_modules_children
        )
        fixed_modules: dict[str, Value] = {
            "typing": StrictBuiltins(typing_children, self.type_env),
            "typing_extensions": StrictBuiltins(
                typing_extensions_children, self.type_env
            ),
            "__strict__": StrictBuiltins(strict_modules_children, self.type_env),
            "strict_modules": StrictBuiltins(
                dict(strict_modules_children), self.type_env
            ),
        }

        builtins_children["<fixed-modules>"] = StrictBuiltins(
            fixed_modules, self.type_env
        )

        self.builtins = self.modules["builtins"] = IntrinsicModuleTable(
            "builtins",
            "<builtins>",
            self,
            builtins_children,
        )
        self.typing = self.modules["typing"] = IntrinsicModuleTable(
            "typing", "<typing>", self, typing_children
        )
        self.modules["typing_extensions"] = ModuleTable(
            "typing_extensions", "<typing_extensions>", self, typing_extensions_children
        )
        self.statics = self.modules["__static__"] = IntrinsicModuleTable(
            "__static__",
            "<__static__>",
            self,
            {
                "Array": self.type_env.array,
                "CheckedDict": self.type_env.checked_dict.exact_type(),
                "CheckedList": self.type_env.checked_list.exact_type(),
                "Enum": self.type_env.enum,
                "IntEnum": self.type_env.int_enum,
                "StringEnum": self.type_env.string_enum,
                "allow_weakrefs": self.type_env.allow_weakrefs,
                "box": BoxFunction(self.type_env.function),
                "cast": CastFunction(self.type_env.function),
                "clen": LenFunction(self.type_env.function, boxed=False),
                "crange": CRangeFunction(self.type_env.function),
                "ExcContextDecorator": self.type_env.exc_context_decorator.exact_type(),
                "ContextDecorator": self.type_env.context_decorator.exact_type(),
                "dynamic_return": self.type_env.dynamic_return,
                "size_t": self.type_env.uint64.exact_type(),
                "ssize_t": self.type_env.int64.exact_type(),
                "cbool": self.type_env.cbool.exact_type(),
                "inline": self.type_env.inline,
                # This is a way to disable the static compiler for
                # individual functions/methods
                "_donotcompile": self.type_env.donotcompile,
                "int8": self.type_env.int8.exact_type(),
                "int16": self.type_env.int16.exact_type(),
                "int32": self.type_env.int32.exact_type(),
                "int64": self.type_env.int64.exact_type(),
                "uint8": self.type_env.uint8.exact_type(),
                "uint16": self.type_env.uint16.exact_type(),
                "uint32": self.type_env.uint32.exact_type(),
                "uint64": self.type_env.uint64.exact_type(),
                "char": self.type_env.char.exact_type(),
                "double": self.type_env.double.exact_type(),
                "unbox": UnboxFunction(self.type_env.function),
                "checked_dicts": self.type_env.bool.instance,
                "prod_assert": ProdAssertFunction(self.type_env),
                "pydict": self.type_env.dict.exact_type(),
                "PyDict": self.type_env.dict.exact_type(),
                "RAND_MAX": rand_max.instance,
                "rand": reflect_builtin_function(
                    # pyre-ignore[6]: Pyre can't know this callable is a BuiltinFunctionType
                    rand,
                    None,
                    self.type_env,
                ),
                "set_type_static": self.type_env.DYNAMIC,
                "native": self.type_env.native_decorator,
                "mixin": self.type_env.function.exact_type(),
                "ClassDecorator": self.type_env.class_decorator.exact_type(),
                "TClass": self.type_env.class_typevar.exact_type(),
            },
        )

        self.modules["dataclasses"] = ModuleTable(
            "dataclasses",
            "Lib/dataclasses.py",
            self,
            {
                "Field": self.type_env.dataclass_field,
                "InitVar": self.type_env.initvar,
                "dataclass": self.type_env.dataclass,
                "field": self.type_env.dataclass_field_function,
            },
        )

        self.modules["cinder"] = self.modules["cinderx"] = ModuleTable(
            "cinder",
            "<cinder>",
            self,
            {
                "cached_property": self.type_env.cached_property,
                "async_cached_property": self.type_env.async_cached_property,
            },
        )

        if xxclassloader is not None:
            spam_obj = self.type_env.spam_obj
            assert spam_obj is not None
            if hasattr(xxclassloader, "foo"):
                funcs = {
                    "foo": reflect_builtin_function(
                        xxclassloader.foo,
                        None,
                        self.type_env,
                    ),
                    "bar": reflect_builtin_function(
                        xxclassloader.bar,
                        None,
                        self.type_env,
                    ),
                    "neg": reflect_builtin_function(
                        xxclassloader.neg,
                        None,
                        self.type_env,
                    ),
                }
            else:
                funcs = {}

            self.modules["xxclassloader"] = ModuleTable(
                "xxclassloader",
                "<xxclassloader>",
                self,
                {
                    "spamobj": spam_obj.exact_type(),
                    "XXGeneric": self.type_env.xx_generic.exact_type(),
                    **funcs,
                },
            )

        self.intrinsic_modules: set[str] = set(self.modules.keys())
        self.decl_visitors: deque[DeclarationVisitor] = deque()

    def __getitem__(self, name: str) -> ModuleTable:
        return self.modules[name]

    def __setitem__(self, name: str, value: ModuleTable) -> None:
        self.modules[name] = value

    def add_module(
        self,
        name: str,
        filename: str,
        tree: AST,
        source: str | bytes | ast.Module | ast.Expression | ast.Interactive,
        optimize: int,
    ) -> ast.Module:
        optimized = AstOptimizer(optimize=optimize > 0).visit(tree)
        assert isinstance(optimized, ast.Module)
        tree = optimized

        self.ast_cache[source] = tree

        # Track if we're the first module being compiled, if we are then
        # we want to finish up the validation of all of the modules that
        # are imported and compiled for the first time because they were
        # imported from this module, this is a little bit odd because we
        # need to finish_bind first before we can validate the overrides
        # and we may have circular references between modules.
        validate_classes = not self.decl_visitors
        decl_visit = DeclarationVisitor(name, filename, self, optimize)
        self.decl_visitors.append(decl_visit)
        decl_visit.visit(tree)
        decl_visit.finish_bind()

        if validate_classes:
            # Validate that the overrides for all of the modules we
            # have compiled from this top-level module.
            while self.decl_visitors:
                decl_visit = self.decl_visitors.popleft()
                decl_visit.module.validate_overrides()

        return tree

    def bind(
        self,
        name: str,
        filename: str,
        tree: AST,
        source: str | bytes,
        optimize: int,
        enable_patching: bool = False,
    ) -> None:
        self._bind(name, filename, tree, source, optimize, enable_patching)

    def _bind(
        self,
        name: str,
        filename: str,
        tree: AST,
        source: str | bytes | ast.Module | ast.Expression | ast.Interactive,
        optimize: int,
        enable_patching: bool = False,
    ) -> tuple[ast.Module, SymbolVisitor]:
        cached_tree = self.ast_cache.get(source)
        if cached_tree is None:
            tree = self.add_module(name, filename, tree, source, optimize)
        else:
            tree = cached_tree
        # Analyze variable scopes
        future_flags = find_futures(0, tree)
        # TASK(TT209531178): This class still implicitly assumes 3.10
        code_generator = cast("Static310CodeGenerator", self.code_generator)
        s = code_generator._SymbolVisitor(future_flags)
        s.visit(tree)

        # Analyze the types of objects within local scopes
        type_binder = self.make_type_binder(
            s,
            filename,
            self,
            name,
            optimize,
            enable_patching=enable_patching,
        )
        type_binder.visit(tree)
        return tree, s

    def make_type_binder(
        self,
        symbols: SymbolVisitor,
        filename: str,
        compiler: Compiler,
        module_name: str,
        optimize: int,
        enable_patching: bool = False,
    ) -> TypeBinder:
        return TypeBinder(
            symbols, filename, compiler, module_name, optimize, enable_patching
        )

    def compile(
        self,
        name: str,
        filename: str,
        tree: AST,
        source: str | bytes,
        optimize: int,
        enable_patching: bool = False,
        builtins: dict[str, Any] = builtins.__dict__,
    ) -> CodeType:
        code_gen = self.code_gen(
            name, filename, tree, source, optimize, enable_patching, builtins
        )
        return code_gen.getCode()

    def code_gen(
        self,
        name: str,
        filename: str,
        tree: AST,
        source: str | bytes | ast.Module | ast.Expression | ast.Interactive,
        optimize: int,
        enable_patching: bool = False,
        builtins: dict[str, Any] = builtins.__dict__,
    ) -> Static310CodeGenerator:
        tree, s = self._bind(name, filename, tree, source, optimize, enable_patching)
        if self.error_sink.has_errors:
            raise self.error_sink.errors[0]

        # Compile the code w/ the static compiler
        graph = self.code_generator.flow_graph(name, filename, s.scopes[tree])
        graph.setFlag(consts.CI_CO_STATICALLY_COMPILED)
        graph.extra_consts.append(tuple(self.modules[name].imported_from.items()))

        future_flags = find_futures(0, tree)

        code_gen = self.code_generator(
            None,
            tree,
            s,
            graph,
            self,
            name,
            flags=0,
            optimization_lvl=optimize,
            enable_patching=enable_patching,
            builtins=builtins,
            future_flags=future_flags,
        )
        code_gen.visit(tree)
        del self.ast_cache[source]
        return cast("Static310CodeGenerator", code_gen)

    def import_module(self, name: str, optimize: int) -> ModuleTable | None:
        pass



# Portions copyright (c) Meta Platforms, Inc. and affiliates.

# pyre-strict

import sys

# operation flags
OP_ASSIGN = "OP_ASSIGN"
OP_DELETE = "OP_DELETE"
OP_APPLY = "OP_APPLY"

SC_LOCAL = 1
SC_GLOBAL_IMPLICIT = 2
SC_GLOBAL_EXPLICIT = 3
SC_FREE = 4
SC_CELL = 5
SC_TYPE_PARAM = 9
SC_UNKNOWN = 6


CO_OPTIMIZED = 0x0001
CO_NEWLOCALS = 0x0002
CO_VARARGS = 0x0004
CO_VARKEYWORDS = 0x0008
CO_NESTED = 0x0010
CO_GENERATOR = 0x0020
CO_NOFREE = 0x0040
CO_COROUTINE = 0x0080
CO_GENERATOR_ALLOWED = 0
CO_ITERABLE_COROUTINE = 0x0100
CO_ASYNC_GENERATOR = 0x0200
CO_FUTURE_DIVISION = 0x20000
CO_FUTURE_ABSOLUTE_IMPORT = 0x40000
CO_FUTURE_WITH_STATEMENT = 0x80000
CO_FUTURE_PRINT_FUNCTION = 0x100000
CO_FUTURE_UNICODE_LITERALS = 0x200000
CO_FUTURE_BARRY_AS_BDFL = 0x400000
CO_FUTURE_GENERATOR_STOP = 0x800000
CO_FUTURE_ANNOTATIONS = 0x1000000
CO_HAS_DOCSTRING = 0x4000000
CO_METHOD = 0x8000000
if sys.version_info >= (3, 14):
    CI_CO_STATICALLY_COMPILED = 0x4000
else:
    CI_CO_STATICALLY_COMPILED = 0x4000000
CO_SUPPRESS_JIT = 0x40000000

PyCF_MASK_OBSOLETE: int = CO_NESTED
PyCF_SOURCE_IS_UTF8 = 0x0100
PyCF_DONT_IMPLY_DEDENT = 0x0200
PyCF_ONLY_AST = 0x0400
PyCF_IGNORE_COOKIE = 0x0800
PyCF_TYPE_COMMENTS = 0x1000
PyCF_ALLOW_TOP_LEVEL_AWAIT = 0x2000
PyCF_COMPILE_MASK: int = (
    PyCF_ONLY_AST
    | PyCF_ALLOW_TOP_LEVEL_AWAIT
    | PyCF_TYPE_COMMENTS
    | PyCF_DONT_IMPLY_DEDENT
)

PyCF_MASK: int = (
    CO_FUTURE_DIVISION
    | CO_FUTURE_ABSOLUTE_IMPORT
    | CO_FUTURE_WITH_STATEMENT
    | CO_FUTURE_PRINT_FUNCTION
    | CO_FUTURE_UNICODE_LITERALS
    | CO_FUTURE_BARRY_AS_BDFL
    | CO_FUTURE_GENERATOR_STOP
    | CO_FUTURE_ANNOTATIONS
    | CI_CO_STATICALLY_COMPILED
)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# pyre-strict

from __future__ import annotations

from asyncio import AbstractEventLoop, CancelledError, Future, get_event_loop
from collections.abc import Awaitable, Callable, Generator
from typing import NoReturn, Optional, TypeVar


try:
    # pyre-ignore[21]: Not including _cinderx as an explicit dependency because this can
    # be used in environments where CinderX is not supported.
    from _cinderx import AsyncLazyValue
except ImportError:

    class _AsyncLazyValueState:
        NotStarted = 0
        Running = 1
        Done = 2

    _TParams = TypeVar("_TParams")
    _T = TypeVar("_T", covariant=True)

    class AsyncLazyValue(Awaitable[_T]):
        """
        This is a low-level class used mainly for two things:
        * It helps to avoid calling a coroutine multiple times, by caching the
        result of a previous call
        * It ensures that the coroutine is called only once

        AsyncLazyValue has well defined cancellation behavior in these cases:

        1. When we have a single task stack (call stack for you JS folks), which is
        awaiting on the AsyncLazyValue
        -> In this case, we mimic the behavior of a normal await. i.e: If the
            task stack gets cancelled, we cancel the coroutine (by raising a
            CancelledError in the underlying future)

        2. When we have multiple task stacks awaiting on the future.
        We have two sub cases here.

        2.1. The initial task stack (which resulted in an await of the coroutine)
                gets cancelled.
                -> In this case, we cancel the coroutine, and all the tasks depending
                on it. If we don't do that, we'd have to implement retry logic,
                which is a bad idea in such low level code. Even if we do implement
                retries, there's no guarantee that they would succeed, so it's better
                to just fail here.

                Also, the number of times this happens is very small (I don't have
                data to prove it, but qualitative arguments suggest this is the
                case).

        2.2. One of the many task stacks gets cancelled (but not the one which ended
            up awaiting the coroutine)
            -> In this case, we just allow the task stack to be cancelled, but
                the rest of them are processed without being affected.
        """

        def __init__(
            self,
            # pyre-fixme[31]: Expression `typing.Callable[(_TParams,
            #  typing.Awaitable[_T])]` is not a valid type.
            coro_func: Callable[_TParams, Awaitable[_T]],
            # pyre-fixme[11]: Annotation `args` is not defined as a type.
            *args: _TParams.args,
            # pyre-fixme[11]: Annotation `kwargs` is not defined as a type.
            **kwargs: _TParams.kwargs,
        ) -> None:
            # pyre-fixme[31]: Expression `typing.Optional[typing.Callable[(_TParams,
            #  typing.Awaitable[_T])]]` is not a valid type.
            self.coro_func: Optional[Callable[_TParams, Awaitable[_T]]] = coro_func
            self.args: tuple[object, ...] = args
            self.kwargs: dict[str, object] = kwargs
            self.state: int = _AsyncLazyValueState.NotStarted
            self.res: Optional[_T] = None
            self._futures: list[Future] = []
            self._awaiting_tasks = 0

        async def _async_compute(self) -> _T:
            futures = self._futures
            try:
                coro_func = self.coro_func
                # lint-fixme: NoAssertsRule
                assert coro_func is not None
                self.res = res = await coro_func(*self.args, **self.kwargs)

                self.state = _AsyncLazyValueState.Done

                # pyre-ignore[1001]: Pyre is worried about how we're accessing futures
                # without awaiting them.
                for value in futures:
                    if not value.done():
                        value.set_result(self.res)

                self.args = ()
                self.kwargs.clear()
                del self._futures[:]
                self.coro_func = None

                return res

            except (Exception, CancelledError) as e:
                # pyre-ignore[1001]: Pyre is worried about how we're accessing futures
                # without awaiting them.
                for value in futures:
                    if not value.done():
                        value.set_exception(e)
                self._futures = []
                self.state = _AsyncLazyValueState.NotStarted
                raise

        def _get_future(self, loop: Optional[AbstractEventLoop]) -> Future:
            if loop is None:
                loop = get_event_loop()
            f = Future(loop=loop)
            self._futures.append(f)
            self._awaiting_tasks += 1
            return f

        def __iter__(self) -> AsyncLazyValue[_T]:
            return self

        def __next__(self) -> NoReturn:
            raise StopIteration(self.res)

        def __await__(self) -> Generator[None, None, _T]:
            if self.state == _AsyncLazyValueState.Done:
                # pyre-ignore[7]: Expected `Generator[None, None, Variable[_T](covariant)]`
                # but got `_AsyncLazyValue[Variable[_T](covariant)]`.
                return self
            elif self.state == _AsyncLazyValueState.Running:
                c = self._get_future(None)
                return c.__await__()
            else:
                self.state = _AsyncLazyValueState.Running
                c = self._async_compute()
                return c.__await__()

        def as_future(self, loop: AbstractEventLoop) -> Future:
            if self.state == _AsyncLazyValueState.Done:
                f = Future(loop=loop)
                f.set_result(self.res)
                return f
            elif self.state == _AsyncLazyValueState.Running:
                return self._get_future(loop)
            else:
                if loop is None:
                    loop = get_event_loop()
                t = loop.create_task(self._async_compute())
                self.state = _AsyncLazyValueState.Running
                # pyre-ignore[16]: Undefined attribute `asyncio.tasks.Task`
                # has no attribute `_source_traceback`.
                if t._source_traceback:
                    del t._source_traceback[-1]
                return t
